/**
 * ViewPort.js - Code to create a viewport window in a web browser.
 *
 * Requires prototypejs 1.5.1+ and scriptaculous 1.7+
 * Requires Array.numericSort(), Array.intersect(), Hash.flip(), and
 * Hash.keyOf() functions (defined in DimpCore).
 *
 * $Horde: dimp/js/src/ViewPort.js,v 1.153 2007/07/27 06:44:17 slusarz Exp $
 *
 * Copyright 2005-2007 The Horde Project (http://www.horde.org/)
 *
 * See the enclosed file COPYING for license information (GPL). If you
 * did not receive this file, see http://www.fsf.org/copyleft/gpl.html.
 */

/**
 * ViewPort
 */
var ViewPort = Class.create();
ViewPort.prototype = {
    // Vars used and defaulting to null/false:
    //   firstcontent, force, id, isbusy, last_request, line_height,
    //   page_size, show_split_pane, splitbar, splitbar_loc, viewport_init
    last_page_size: -1,
    last_start_pos: -1,

    // Required options: content_container, lines, fetch_action, template,
    //                   cachecheck_action, ajaxRequest, buffer_pages,
    //                   limit_factor, content_class, row_class, maxHeight,
    //                   msgid_prefix
    initialize: function(opts) {
        opts.content = $(opts.content_container);
        opts.empty = opts.empty_container ? $(opts.empty_container) : null;
        opts.error = opts.error_container ? $(opts.error_container) : null;
        this.opts = opts;

        this.scroller = new ViewPort_Scroller(this);
        this.template = new Template(opts.template)
        this.views = $H();
    },

    loadView: function(id, params, uid) {
        var init = false;
        this.clearWait();

        if (this.metadata) {
            this.views[this.id] = { metadata: this.metadata, buffer: this.buffer, offset: this.currentOffset() };
            if (this.opts.onViewChange) {
                this.opts.onViewChange();
            }
        } else {
            init = true;
        }

        this.additional_params = $H(params);
        this.id = id;

        if (this.views[id]) {
            if (this.filter) {
                this.filter.reset();
            }
            this.metadata = this.views[id].metadata;
            this.buffer = this.views[id].buffer;
            this.updateContent(this.views[id].offset);
            this.checkCache();
            if (this.opts.onComplete) {
                this.opts.onComplete();
            }
            return true;
        }

        this.metadata = new ViewPort_MetaData();
        this.buffer = new ViewPort_Buffer(this, this.opts.buffer_pages, this.opts.limit_factor);

        if (init) {
            // If the first time we are loading the viewport, and no default
            // page size has been given, try to determine a default size.
            if (this.page_size == undefined) {
                this.page_size = (this.show_split_pane) ? this.getDefaultPageSize() : this.getMaxPageSize();
            }
            this.viewport_init = true;
            this.renderViewport();
        }

        this.fetchBuffer((uid) ? 'uid' : 'offset', (uid) ? uid : 0);
        this.clearWait();
        return false;
    },

    deleteView: function(id) {
        this.views.remove(id);
    },

    getRowById: function(id) {
        return (this.buffer) ? this.buffer.getRowById(id) : null;
    },

    inBuffer: function(id) {
        return this.buffer.inBuffer(id);
    },

    positionById: function(id) {
        var row = this.getRowById(id);
        if (!row) {
            return false;
        }

        switch (this.visible(id)) {
        case -1:
            this.scroller.moveScroll(row.rownum - 1);
            break;

        // case 0:
        // noop

        case 1:
            this.scroller.moveScroll(row.rownum - this.getPageSize());
        }

        return true;
    },

    visible: function(id) {
        var row = this.getRowById(id);
        if (!row) {
            return false;
        }

        if (row.rownum < this.last_start_pos + 1) {
            return -1;
        } else if (row.rownum > (this.last_start_pos + this.getCurrPageSize())) {
            return 1;
        } else {
            return 0;
        }
    },

    reload: function(params) {
        if (this.filter && this.filter.isFiltering()) {
            this.filter.filter(true);
        } else {
            this.fetchBuffer('offset', this.currentOffset(), null, $H(params).merge({ update: 1 }));
        }
    },

    refresh: function() {
        if (!this.requestContentRefresh(this.currentOffset()) &&
            this.opts.onComplete) {
            // Even if there has not been a change to the current view, other
            // viewport parameters may have changed so we need to call
            // onComplete() to give other elements a chance to refresh.
            this.opts.onComplete();
        }
    },

    remove: function(uids, cacheid) {
        this.setMetaData({ cacheid: cacheid });

        var visible = uids.intersect(this.buffer.getRows(this.last_start_pos + 1, this.last_page_size).pluck('uid'));

        // If we have visible elements to remove, only call refresh/reload
        // after the last effect has finished.
        this.isbusy = false;
        var vsize = visible.size();
        if (vsize) {
            if (this.opts.onClearRows) {
                this.opts.onClearRows(visible);
            }
            for (var i = 0; i < vsize; ++i) {
                this.isbusy = true;
                var args = { duration: 0.4 };
                if ((i + 1) == vsize) {
                    args.afterFinish = this.removeids.bind(this, uids);
                }
                Effect.Fade(this.getRowDiv(visible[i]), args);
            }
            this.last_page_size -= vsize;
        } else {
            this.removeids(uids);
        }
    },

    removeids: function(uids) {
        if (this.opts.onRemoveRows) {
            this.opts.onRemoveRows(uids.collect(function(id) {
                return this.getRowById(id);
            }.bind(this)).compact());
        }

        this.buffer.removeByUid(uids);
        this.setMetaData({ total_rows: this.getMetaData('total_rows') - uids.size() });
        this.refresh();
        this.isbusy = false;
    },

    addFilter: function(input, action, callback) {
        this.filter = new ViewPort_Filter(this, input, action, callback);
    },

    runFilter: function() {
        if (this.filter) {
            this.filter.filter(true);
        }
    },

    onResize: function() {
        if (!this.firstcontent || !this.opts.content.visible()) {
            return;
        }

        if (this.opts.onBeforeResize) {
            this.opts.onBeforeResize();
        }

        this.renderViewport();

        if (this.opts.onAfterResize) {
            this.opts.onAfterResize();
        }
    },

    requestContentRefresh: function(offset) {
        if (offset == this.last_start_pos &&
            this.getPageSize() == this.last_page_size) {
            return false;
        }
        if (this.buffer.isLoaded(offset)) {
            this.updateContent(offset);
            var limit = this.buffer.isNearingLimit(offset);
            if (limit) {
                this.fetchBuffer('offset', offset, { noUpdateContent: offset }, { nearing: limit, page_size: this.getPageSize() });
            }
            return true;
        } else {
            this.fetchBuffer('offset', offset);
            return false;
        }
    },

    fetchBuffer: function(type, value, opts, params) {
        if (this.opts.onFetch) {
            this.opts.onFetch();
        }

        params = this.addRequestParams(params);
        params[type] = value;
        var action = this.opts.fetch_action;

        if (this.filter && this.filter.isFiltering()) {
            action = this.filter.getAction();
            params = this.filter.addFilterParams(params);
        }

        var lr_count = 1;
        if (this.last_request &&
            // This means the previous request is still processing.
            this.last_request.count &&
            this.last_request.type == type &&
            this.last_request.id == this.id) {
            if (this.last_request.value == value) {
                // Check for repeat request.  We technically should never
                // reach here but if we do, make sure we don't go into an
                // infinite loop.
                if (++this.last_request.count == 4) {
                    this.displayFetchError();
                    return;
                }
                lr_count = this.last_request.count;
            } else if (this.last_request.type == 'offset' &&
                       value > this.last_request.value &&
                       ((value + this.getPageSize()) < (this.last_request.value + this.last_request.size))) {
                // Check for message list requests that are requesting
                // (essentially) the same message slice - such as two
                // scroll down requests sent in quick succession.  If the
                // original request will contain the slice needed by the
                // second request, ignore the later request.
                this.last_request.offset = value;
                return;
            }
        }

        this.last_request = { type: type, value: value, id: this.id, size: params.buffer_size, count: 1, offset: null };

        this.opts.ajaxRequest(action, params, function(r) { this.ajaxUpdate(r, { offset: ((type == 'offset') ? value : null), options: opts || $H(), id: this.id }) }.bind(this), Object.clone(this.opts));
        this.handleWait();
    },

    addRequestParams: function(args) {
        var params = Object.clone(this.additional_params),
            cid = this.getMetaData('cacheid');
        if (cid) {
            params = $H(params).merge({ cacheid: cid });
        }
        return params.merge(args).merge({ buffer_size: this.buffer.bufferSize(), offset: this.currentOffset() });
    },

    ajaxUpdate: function(r, opts) {
        this.clearWait();

        if (!r.response || !r.response.rows) {
            if (this.id == opts.id) {
                this.displayFetchError();
            }
            return;
        }

        r.response.viewport_done = true;

        var d = r.response.rows;
        var showing = (this.id == d.id || d.search);
        var buffer, metadata, view;

        if (showing) {
            buffer = this.buffer;
            metadata = this.metadata;
        } else {
            view = this.views[d.id];
            buffer = view.buffer;
            metadata = view.metadata;
            view.offset = opts.offset;
        }

        // Parse parameters
        d.msglist = $H(d.msglist);
        buffer.update(d.msglist, d.msgdata, d.update);
        metadata.set($H(d.other).merge({
            cacheid: d.cacheid,
            label: d.label,
            total_rows: d.totalrows
        }));

        // Don't update the view if we are now in a different view, or if
        // the offset has changed since we originally requested the chunk.
        if (!showing ||
            (this.last_request.type == 'offset' && d.offset != this.last_request.value) ||
            (this.last_request.type == 'uid' && d.uid != this.last_request.value)) {
            return;
        }

        if (!opts.options.noUpdateContent ||
            (this.last_request.count &&
             opts.options.noUpdateContent != this.currentOffset())) {
            var offset;
            if (this.last_request.offset != undefined) {
                offset = this.last_request.offset;
            } else if (opts.offset == null) {
                // If offset is null, we originally searched by ID. Thus, we
                // need to determine the offset by what was returned from the
                // server.
                offset = d.msglist.keyOf(d.uid);
                offset = (offset === undefined) ? 0 : (offset - 1);
            } else {
                // If we delete off the bottom of the page, our offset won't
                // account for this and we will have to do 2 updateContent()
                // calls - 1 now and 1 in scrollIdle().  Catch this case and
                // adjust the offset accordingly.
                offset = Math.min(opts.offset, this.getMetaData('total_rows') - this.getCurrPageSize());
            }
            if (!this.updateContent(offset)) {
                return;
            }
        }

        // Set last_request.count to 0.  This indicates that the last request
        // has finished loading.
        this.last_request.count = 0;

        if (this.opts.onComplete) {
            this.opts.onComplete();
        }
    },

    updateContent: function(offset) {
        if (!this.buffer.isLoaded(offset)) {
            this.fetchBuffer('offset', offset);
            return false;
        }

        this.last_start_pos = offset;
        this.last_page_size = this.getPageSize();
        var c = this.opts.content, c_ids = [], c_remove = [], c_rows = [];

        // Code for viewport that only needs to be initialized once.
        if (!this.firstcontent) {
            this.firstcontent = true;
            if (this.opts.onFirstContent) {
                this.opts.onFirstContent();
            }
        } else {
            c_rows = c.immediateDescendants();
            c_ids = c_rows.pluck('id');
        }

        var newrows = [],
            rows = this.getRows(offset + 1, this.last_page_size);

        c_rows.invoke('remove');

        $R(0, rows.size() - 1).each(function(i) {
            var r;
            if (!this.force) {
                var key = c_ids.indexOf(this.opts.msgid_prefix + rows[i].uid);
                if (key != -1) {
                    r = c_rows[key];
                    c_remove.push(key);
                }
            }
            if (!r) {
                r = this.buffer.createRowDiv(rows[i]);
                newrows.push(r);
            }
            c.appendChild(r);
        }.bind(this));

        c_remove.each(function(k) {
            c_rows[k] = null;
        });
        c_rows = c_rows.compact();

        if (c_rows.size() && this.opts.onClearRows) {
            this.opts.onClearRows(c_rows);
        }

        if (newrows.size() && this.opts.onContent) {
            this.opts.onContent(newrows);
        }

        if (this.opts.empty && !c.immediateDescendants().size()) {
            c.update(this.opts.empty.innerHTML);
        }

        this.scroller.updateSize();
        this.scroller.moveScroll(offset);

        return true;
    },

    // If true, forces an update of all visible rows on viewport redraw.
    forceUpdate: function(force) {
        this.force = force;
    },

    displayFetchError: function() {
        if (this.opts.onFail) {
            this.opts.onFail();
        }
        if (this.opts.error) {
            this.opts.content.update(this.opts.error.innerHTML);
        }
    },

    checkCache: function() {
        this.opts.ajaxRequest(this.opts.cachecheck_action, this.addRequestParams($H()), this.checkCacheCallback.bind(this));
    },

    checkCacheCallback: function(r) {
        // No need to do anything if ajaxUpdate() has already processed the
        // data.
        if (!r.response || !r.response.rows || r.response.viewport_done) {
            return;
        }
        if (this.isbusy) {
            setTimeout(this.checkCacheCallback.bind(this, r), 100);
        } else {
            this.last_request = { type: 'cache' };
            this.ajaxUpdate(r, { offset: this.currentOffset(), id: this.id, options: $H() });
        }
    },

    handleWait: function(call) {
        this.clearWait();

        // Server did not respond in defined amount of time.  Alert the
        // callback function and set the next timeout.
        if (call && this.opts.onWait) {
            this.opts.onWait();
        }

        // Call wait handler every x seconds
        if (this.opts.viewport_wait) {
            this.waitHandler = setTimeout(function() { this.handleWait(true); }.bind(this), this.opts.viewport_wait * 1000);
        }
    },

    clearWait: function() {
        if (this.waitHandler) {
            clearTimeout(this.waitHandler);
            this.waitHandler = null;
        }
    },

    clearMsgList: function() {
        var c = this.opts.content;

        if (this.opts.onClearRows) {
            this.opts.onClearRows(c.immediateDescendants());
        }

        $A(c.childNodes).each(function(n) {
            c.removeChild(n);
        });
    },

    clear: function() {
        this.clearMsgList();
        this.scroller.clear();
    },

    visibleRows: function() {
        return this.opts.content.immediateDescendants();
    },

    getRows: function(start, length) {
        return this.buffer.getRows(start, length);
    },

    getRowDiv: function(id) {
        var row = this.getRowById(id);
        return (row) ? $(this.opts.msgid_prefix + row.uid) : null;
    },

    getMetaData: function(id) {
        return this.metadata.get(id);
    },

    setMetaData: function(vals, id) {
        var md;
        if (id && id != this.id) {
            if (!this.views[id]) {
                return;
            }
            md = this.views[id].metadata;
        } else {
            md = this.metadata;
        }
        md.set(vals);
    },

    currentOffset: function() {
        return this.scroller.currentOffset();
    },

    updateFlag: function(ids, flag, add, permanent) {
        ids.each(function(id) {
            var row = this.getRowById(id);
            if (!row) {
                return;
            }

            if (permanent) {
                row.bg = (add) ? $w(row.bg).concat([ flag ]).join(' ') : $w(row.bg).without([ flag ]).join(' ');
            }

            row = this.getRowDiv(id);
            if (row) {
                if (add) {
                    row.addClassName(flag);
                } else {
                    row.removeClassName(flag);
                }
            }
        }.bind(this));
    },

    getBufferSize: function() {
        return this.buffer.size;
    },

    getLineHeight: function() {
        if (this.line_height) {
            return this.line_height;
        }

        // To avoid hardcoding the line height, create a temporary row to
        // figure out what the CSS says.
        var d = $(Builder.node('DIV', { className: this.opts.content_class }, [ Builder.node('DIV', { className: this.opts.row_class }) ])).hide();
        document.body.appendChild(d);
        this.line_height = d.getHeight();
        document.body.removeChild(d);

        return this.line_height;
    },

    setPageSize: function(size) {
        this.page_size = size;
        this.renderViewport();
    },

    getPageSize: function() {
        return this.page_size;
    },

    getCurrPageSize: function() {
        return Math.min(this.getPageSize(), this.getMetaData('total_rows'));
    },

    getDefaultPageSize: function() {
        return Math.max(parseInt(this.getMaxPageSize() * 0.45), 5);
    },

    getMaxPageSize: function() {
        return parseInt(this.opts.maxHeight() / this.getLineHeight());
    },

    showSplitPane: function(show) {
        this.show_split_pane = show;
        this.onResize();
    },

    renderViewport: function() {
        if (!this.viewport_init) {
            return;
        }

        var lh = this.getLineHeight(), panesize;

        // Get split pane dimensions
        if (this.opts.split_pane) {
            var pane = $(this.opts.split_pane);
            if (this.show_split_pane) {
                if (!pane.visible()) {
                    this.initSplitBar();
                    $(pane, this.splitbar).invoke('show');
                    this.page_size = (this.splitbar_loc) ? this.splitbar_loc : this.getDefaultPageSize();
                }
                panesize = this.opts.maxHeight() - (lh * (this.page_size + 1));
            } else if (pane.visible()) {
                this.splitbar_loc = this.page_size;
                $(pane, this.splitbar).invoke('hide');
                this.page_size = this.getMaxPageSize();
            }
        }

        // Size the panes
        $(this.opts.content).setStyle({ height: (lh * this.page_size) + 'px' });
        if (panesize) {
            pane.setStyle({ height: panesize + 'px' });
        }

        this.scroller.onResize();
    },

    initSplitBar: function() {
        if (this.splitbar) {
            return;
        }

        this.splitbar = $(this.opts.splitbar);
        new Draggable(this.opts.splitbar, {
            constraint: 'vertical',
            ghosting: true,
            onStart: function() {
                // Cache these values since we will be using them multiple
                // times in snap().
                var lh = this.getLineHeight();
                this.sp = { lh: lh, pos: Position.positionedOffset($(this.opts.content))[1], max: parseInt((this.opts.maxHeight() - 100) / lh) };
            }.bind(this),
            snap: function(x, y) {
                var l = parseInt((y - this.sp.pos) / this.sp.lh);
                if (l < 1) {
                    l = 1;
                } else if (l > this.sp.max) {
                    l = this.sp.max;
                }
                this.sp.lines = l;
                return [0, this.sp.pos + (l * this.sp.lh)];
            }.bind(this),
            onEnd: function() {
                $(this.opts.splitbar).setStyle({ top: 0 });
                this.setPageSize(this.sp.lines);
            }.bind(this)
        });
    }

};

/**
 * ViewPort_MetaData
 *
 * Required metadata: cacheid, label, total_rows
 */
var ViewPort_MetaData = Class.create();
ViewPort_MetaData.prototype = {

    total_rows: 0,

    initialize: function() {
    },

    get: function(id) {
        return (this[id] != undefined) ? this[id] : null;
    },

    set: function(vals) {
        $H(vals).each(function(node) {
            this[node.key] = node.value;
        }.bind(this));
    }

};

/**
 * ViewPort_Scroller
 */
var ViewPort_Scroller = Class.create();
ViewPort_Scroller.prototype = {
    // Vars used and defaulting to null:
    //   scroll_event

    initialize: function(parent) {
        this.parent = parent;
    },

    createScrollBar: function() {
        if (this.scrollerDiv) {
            return false;
        }
        var c = this.parent.opts.content;

        // Create the outer div.
        this.scrollerDiv = $(Builder.node('DIV', { className: 'sbouterdiv', id: 'sbouterdiv', style: 'height:' + c.getHeight() + 'px;' }));
        this.scrollerDiv.observe('scroll', this.handleScroll.bindAsEventListener(this));

        // Give our parent a right margin just big enough to
        // accomodate the scrollbar.
        c.setStyle({ marginRight: '-19px' });

        // Create the inner div.
        this.heightDiv = $(Builder.node('DIV', { className: 'sbinnerdiv', id: 'sbinnerdiv' }));
        this.updateSize();

        this.scrollerDiv.appendChild(this.heightDiv);
        c.parentNode.insertBefore(this.scrollerDiv, c.nextSibling);

        // Mouse wheel handler.
        c.observe(Prototype.Browser.IE ? 'mousewheel' : 'DOMMouseScroll', function(e) {
            this.scrollerDiv.scrollTop += (e.wheelDelta >= 0 || e.detail < 0)
                // wheel-up.
                ? (-3 * this.parent.getLineHeight())
                : (3 * this.parent.getLineHeight());
            this.onScroll();
            if (this.parent.opts.onMouseScroll) {
                this.parent.opts.onMouseScroll(e);
            }
            Event.stop(e);
        }.bindAsEventListener(this), false);

        return true;
    },

    onResize: function() {
        if (!this.createScrollBar()) {
            // Update the outer div.
            this.scrollerDiv.style.height = this.parent.opts.content.getHeight() + 'px';

            // Update the inner div.
            this.updateSize();
        }

        // Update displayed content.
        this.parent.refresh();
        this.onScroll();
    },

    updateSize: function() {
        if (this.heightDiv) {
            this.heightDiv.setStyle({ height: Math.round(this.parent.opts.content.getHeight() * this.parent.getMetaData('total_rows') / this.parent.getPageSize()) + 'px' });
        }
    },

    clear: function() {
        if (this.scrollerDiv) {
            var h = this.parent.opts.content.getHeight() + 'px';
            this.scrollerDiv.setStyle({ height: h });
            this.heightDiv.setStyle({ height: h });
        }
    },

    moveScroll: function(offset) {
        if (this.scrollerDiv) {
            this.scrollerDiv.scrollTop = Math.round((offset / this.parent.getMetaData('total_rows')) * this.heightDiv.getHeight());
            this.onScroll();
        }
    },

    onScroll: function() {
        if (this.parent.opts.onScroll) {
            this.parent.opts.onScroll();
        }

        this.parent.requestContentRefresh(this.currentOffset());

        if (this.parent.opts.onScrollIdle) {
            this.parent.opts.onScrollIdle();
        }
    },

    handleScroll: function(e) {
        if (this.scroll_event) {
            return;
        }
        this.scroll_event = true;
        setTimeout(function() {
            this.onScroll();
            this.scroll_event = false;
        }.bind(this), 50);
    },

    currentOffset: function() {
        if (!this.scrollerDiv) {
            return 0;
        }
        var rows = this.parent.getMetaData('total_rows');
        var offset = Math.min(rows, Math.round(this.scrollerDiv.scrollTop * (rows / this.heightDiv.getHeight())));
        return isNaN(offset) ? 0 : offset;
    }

};

/**
 * ViewPort_Buffer
 *
 * Note: recognize the difference between offset (current location in the
 * viewport - starts at 0) with start parameters (the row numbers - starts
 * at 1).
 *
 * Also, due to limitations in PHP casting an array to an object, the keys
 * of this.list will be strings.  Always convert to integer before using.
 * We don't do this in PHP since casting from an array to an object is much
 * less memory/resource intensive than looping through the array.
 */
var ViewPort_Buffer = Class.create();
ViewPort_Buffer.prototype = {

    initialize: function(parent, b_pages, l_factor) {
        this.clear();
        this.bufferPages = b_pages;
        this.limitFactor = l_factor;
        this.parent = parent;
    },

    limitTolerance: function() {
        return Math.round(this.bufferSize() * (this.limitFactor / 100));
    },

    bufferSize: function() {
        return Math.round(this.bufferPages * this.parent.getPageSize());
    },

    update: function(list, data, update) {
        this.data = this.data.merge(data);
        this.list = (update) ? $H(list) : this.list.merge(list);
        this.size = this.list.size();
        this.lookup = this.list.flip();
    },

    isLoaded: function(offset) {
        return $R(offset + 1, Math.min(offset + this.parent.getPageSize() - 1, this.parent.getMetaData('total_rows'))).all(function(o) {
            return this.list[o] != undefined;
        }.bind(this));
    },

    isNearingTopLimit: function(offset) {
        if (offset == 0) {
            return false;
        }
        return $R(Math.max(offset + 1 - this.limitTolerance(), 1), offset).any(function(o) {
            return this.list[o] == undefined;
        }.bind(this));
    },

    isNearingBottomLimit: function(offset) {
        // Search for missing messages in reverse order since in normal usage
        // (sequential scrolling through the message list) messages are
        // more likely to be missing at furthest from the current view.
        return $A($R(offset + 1, Math.min(offset + this.limitTolerance() + this.parent.getPageSize() - 1, this.parent.getMetaData('total_rows')))).reverse().any(function(o) {
            return this.list[o] == undefined;
        }.bind(this));
    },

    isNearingLimit: function(offset) {
        if (this.size != this.parent.getMetaData('total_rows')) {
            if (this.isNearingTopLimit(offset)) {
                return 'top';
            } else if (this.isNearingBottomLimit(offset)) {
                return 'bottom';
            }
        }
        return false;
    },

    getRowById: function(id) {
        return (this.lookup[id]) ? this.getRows(parseInt(this.lookup[id]), 1).first() : null;
    },

    getRows: function(start, length) {
        return $R(start, Math.min(start + length - 1, this.parent.getMetaData('total_rows'))).collect(function(num) {
            if (this.list[num] != undefined) {
                var uid = this.list[num];
                // We can directly write the rownum to the original object
                // since we will always rewrite when creating rows.
                this.data[uid].rownum = num;
                return this.data[uid];
            }
        }.bind(this)).compact();
    },

    inBuffer: function(id) {
        return this.lookup[id];
    },

    createRowDiv: function(row) {
        return $(Builder.node('DIV')).update(this.parent.template.evaluate(row)).down();
    },

    removeByUid: function(uids) {
        var newlist = $H(), rowsubtract = 0, uidremove = [];
        this.list.keys().numericSort().each(function(n) {
            var uid = this.list[n];
            var key = uids.indexOf(uid);
            if (key == -1) {
                var newrow = n - rowsubtract;
                newlist[newrow] = uid;
                this.lookup[uid] = newrow;
            } else {
                uidremove.push(uid);
                ++rowsubtract;
            }
        }.bind(this));
        [ this.data, this.lookup ].invoke('remove', uidremove);
        this.list = newlist;
        this.size = this.list.size();
    },

    clear: function() {
        this.data = $H();
        this.list = $H();
        this.lookup = $H();
        this.size = 0;
    }

};

/**
 * ViewPort_Filter
 */
var ViewPort_Filter = Class.create();
ViewPort_Filter.prototype = {
    // Vars used and defaulting to null/false:
    //   filtering, last_filter, last_folder, last_folder_params, observer
    filterid: 0,

    initialize: function(parent, input, action, callback) {
        this.parent = parent;
        this.input = $(input);
        this.action = action;
        this.callback = callback;

        this.input.observe('keyup', this.onKeyUp.bind(this));
    },

    onKeyUp: function() {
        if (this.observer) {
            clearTimeout(this.observer);
        }
        this.observer = setTimeout(this.filter.bind(this), 500);
    },

    filter: function(force) {
        var val = $F(this.input).toLowerCase();
        if (!val) {
            this.clear();
            return;
        }

        if (!force && (val == this.last_filter)) {
            return;
        }
        this.last_filter = val;

        if (this.filtering) {
            this.parent.fetchBuffer('offset', 0);
            return;
        }

        this.filtering = ++this.filterid + '%search%';
        this.last_folder = this.parent.id;
        this.last_folder_params = this.parent.additional_params;

        // Filter visible rows immediately.
        var c = this.parent.opts.content;
        var delrows = c.immediateDescendants().findAll(function(n) {
            return n.getText(true).toLowerCase().indexOf(val) == -1;
        });
        if (this.parent.opts.onClearRows) {
            this.parent.opts.onClearRows(delrows);
        }
        delrows.invoke('remove');
        this.parent.scroller.clear();
        if (this.parent.opts.empty && !c.immediateDescendants().size()) {
            c.update(this.parent.opts.empty.innerHTML);
        }

        this.parent.loadView(this.filtering, Object.clone(this.parent.additional_params));
    },

    isFiltering: function() {
        return this.filtering;
    },

    getAction: function() {
        return this.action;
    },

    addFilterParams: function(params) {
        if (!this.filtering) {
            return params;
        }

        params = params.merge({ filter: this.last_filter });

        // Get parameters from a callback function, if defined.
        if (this.callback) {
            params = params.merge($H(this.callback()));
        }

        return params;
    },

    reset: function() {
        if (!this.filtering) {
            return;
        }
        this.parent.deleteView(this.filtering);
        this.filtering = this.last_filter = this.last_folder = null;
    },

    clear: function() {
        if (this.filtering) {
            this.parent.loadView(this.last_folder, this.last_folder_params);
            this.clear();
        }
    }

};
